// Copyright (C) 1998 Bruce Guenter <bruce.guenter@qcc.sk.ca>
//
// This package is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

#include "mystring.h"
#include <ctype.h>

#ifdef TRACE
#include <iostream.h>
ostream& operator<<(ostream& out, const mystringtmp& s)
{
  out << s;
  return out;
}

#define trace(X) cerr << (void*)this << "->" << __PRETTY_FUNCTION__ << X << endl
#define trace_static(X) cerr << __PRETTY_FUNCTION__ << X << endl
#else
#define trace(X) do { } while(0)
#define trace_static(X) do { } while(0)
#endif

mystringrep nil = { 0, 1, "" };

static const unsigned replength = sizeof(unsigned)*2;

///////////////////////////////////////////////////////////////////////////////
// class mystringrep
///////////////////////////////////////////////////////////////////////////////
mystringrep* mystringrep::alloc(unsigned length)
{
  trace_static("length=" << length);
  if(length == 0)
    return &nil;
  unsigned size = length + replength+1;
  mystringrep* ptr = (mystringrep*)new char[size];
  ptr->length = length;
  ptr->references = 0;
  return ptr;
}

mystringrep* mystringrep::dup(const char* str, unsigned length)
{
  trace_static("str=" << (void*)str << " length=" << length);
  if(length == 0)
    return &nil;
  mystringrep* ptr = alloc(length);
  memcpy(ptr->buf, str, length);
  ptr->buf[length] = 0;
  return ptr;
}

mystringrep* mystringrep::dup(const char* str1, unsigned length1,
			      const char* str2, unsigned length2)
{
  trace_static("");
  if(length1+length2 == 0)
    return &nil;
  mystringrep* ptr = alloc(length1+length2);
  memcpy(ptr->buf, str1, length1);
  memcpy(ptr->buf+length1, str2, length2);
  ptr->buf[length1+length2] = 0;
  return ptr;
}

void mystringrep::attach()
{
  trace("references=" << references);
  ++references;
}

void mystringrep::detach()
{
  trace("references=" << references);
  
  --references;
  if(!references) {
    trace("deleting this");
    delete this;
  }
}

///////////////////////////////////////////////////////////////////////////////
// class mystringtmp
///////////////////////////////////////////////////////////////////////////////
void mystringtmp::checkstr()
{
  // If the string is NULL or the length is zero or
  // the length is unset and the string is empty:
  // set the representation to nil
  if(!str || !len || (len == badlen && !str[0])) {
    rep = &nil;
    str = rep->buf;
    len = 0;
  }
}

unsigned mystringtmp::length()
{
  checkstr();
  if(len == badlen)
    len = strlen(str);
  return len;
}

mystringrep* mystringtmp::getrep()
{
  if(!rep)
    rep = mystringrep::dup(str, length());
  return rep;
}

mystringtmp::operator const char*()
{
  checkstr();
  return str;
}

bool mystringtmp::operator!()
{
  checkstr();
  return len == 0;
}

mystringtmp::mystringtmp(const mystringtmp& in)
  : rep(in.rep), str(in.str), len(in.len)
{
}

mystringtmp::mystringtmp(const char* s)
  : rep(0), str(s), len(badlen)
{
}

mystringtmp::mystringtmp(const char* s, unsigned l)
  : rep(0), str(s), len(l)
{
}
mystringtmp::mystringtmp(mystringrep* r)
  : rep(r), str(r->buf), len(r->length)
{
}

mystringtmp::mystringtmp(const mystring& s)
  : rep(s.rep), str(s.rep->buf), len(s.rep->length)
{
}
			
///////////////////////////////////////////////////////////////////////////////
// class mystring
///////////////////////////////////////////////////////////////////////////////
mystring::~mystring()
{
  trace("rep=" << (void*)rep);
  rep->detach();
}

void mystring::dupnil()
{
  trace("");
  rep = &nil;
  rep->attach();
}

void mystring::dup(mystringtmp in)
{
  trace("in='" << in << "'");
  rep = in.getrep();
  rep->attach();
}

void mystring::append(mystringtmp str)
{
  trace("");
  if(!str)
    return;
  if(!*this)
    assign(str);
  else {
    mystringrep* newrep = mystringrep::dup(rep->buf, rep->length,
					   str, str.length());
    rep->detach();
    rep = newrep;
    rep->attach();
  }
}

void mystring::assign(mystringtmp in)
{
  trace("in='" << in << "'");
  if(in.str != rep->buf) {
    mystringrep* tmp = rep;
    dup(in);
    tmp->detach();
  }
}

mystring::mystring(const char* s)
{
  dup(s);
}

mystring::mystring(const mystring& s)
{
  dup(s);
}

mystring::mystring(const char* str, size_t len)
{
  dup(mystringtmp(str, len));
}

mystring::mystring(mystringtmp x, mystringtmp y)
{
  trace("");
  if(!x) dup(y);
  else if(!y) dup(x);
  else {
    rep = mystringrep::dup(x, x.length(), y, y.length());
    rep->attach();
  }
}

mystring::mystring(mystringtmp a, mystringtmp b, mystringtmp c)
{
  trace("");
  rep = mystringrep::alloc(a.length()+b.length()+c.length());
  rep->attach();
  char* ptr = rep->buf;
  memcpy(ptr, a, a.length()); ptr += a.length();
  memcpy(ptr, b, b.length()); ptr += b.length();
  memcpy(ptr, c, c.length()); ptr += c.length();
  *ptr = 0;
}

mystring::mystring(mystringtmp a, mystringtmp b, mystringtmp c, mystringtmp d)
{
  trace("");
  rep = mystringrep::alloc(a.length()+b.length()+c.length()+d.length());
  rep->attach();
  char* ptr = rep->buf;
  memcpy(ptr, a, a.length()); ptr += a.length();
  memcpy(ptr, b, b.length()); ptr += b.length();
  memcpy(ptr, c, c.length()); ptr += c.length();
  memcpy(ptr, d, d.length()); ptr += d.length();
  *ptr = 0;
}

mystring::mystring(mystringtmp a, mystringtmp b, mystringtmp c,
		   mystringtmp d, mystringtmp e)
{
  trace("");
  rep = mystringrep::alloc(a.length()+b.length()+c.length()+d.length()+e.length());
  rep->attach();
  char* ptr = rep->buf;
  memcpy(ptr, a, a.length()); ptr += a.length();
  memcpy(ptr, b, b.length()); ptr += b.length();
  memcpy(ptr, c, c.length()); ptr += c.length();
  memcpy(ptr, d, d.length()); ptr += d.length();
  memcpy(ptr, e, e.length()); ptr += e.length();
  *ptr = 0;
}

mystring::mystring(mystringtmp a, mystringtmp b, mystringtmp c,
		   mystringtmp d, mystringtmp e, mystringtmp f)
{
  trace("");
  rep = mystringrep::alloc(a.length()+b.length()+c.length()+d.length()+e.length()+f.length());
  rep->attach();
  char* ptr = rep->buf;
  memcpy(ptr, a, a.length()); ptr += a.length();
  memcpy(ptr, b, b.length()); ptr += b.length();
  memcpy(ptr, c, c.length()); ptr += c.length();
  memcpy(ptr, d, d.length()); ptr += d.length();
  memcpy(ptr, e, e.length()); ptr += e.length();
  memcpy(ptr, f, f.length()); ptr += f.length();
  *ptr = 0;
}

int mystring::operator!=(const char* in) const
{
  if(rep->buf == in)
    return 0;
  return strcmp(rep->buf, in);
}

int mystring::operator!=(const mystring& in) const
{
  if(rep->buf == in.rep->buf)
    return 0;
  return strcmp(rep->buf, in.rep->buf);
}

mystring mystring::operator=(const char* in)
{
  assign(in);
  return *this;
}

mystring mystring::operator=(const mystring& in)
{
  assign(in);
  return *this;
}
  

mystring mystring::subst(char from, char to) const
{
  const unsigned length = rep->length;
  char buf[length+1];
  const char* in = rep->buf + length;
  for(char* out = buf+length; out >= buf; in--, out--)
    *out = (*in == from) ? to : *in;
  return mystring(buf, length);
}

mystring mystring::lower() const
{
  const unsigned length = rep->length;
  char buf[length+1];
  const char* in = rep->buf + length;
  for(char* out = buf+length; out >= buf; in--, out--)
    *out = isupper(*in) ? tolower(*in) : *in;
  return mystring(buf, length);
}

mystring mystring::upper() const
{
  const unsigned length = rep->length;
  char buf[length+1];
  const char* in = rep->buf + length;
  for(char* out = buf+length; out >= buf; in--, out--)
    *out = islower(*in) ? toupper(*in) : *in;
  return mystring(buf, length);
}

int mystring::find(char ch, size_t offset) const
{
  if(offset >= rep->length)
    return -1;
  char* ptr = strchr(rep->buf+offset, ch);
  return ptr ? ptr-rep->buf : -1;
}

int mystring::find_last(char ch) const
{
  char* ptr = strrchr(rep->buf, ch);
  return ptr ? ptr-rep->buf : -1;
}
  
int mystring::find_first(mystringtmp set, size_t offset) const
{
  const char* setstr = set;
  for(; offset < rep->length; offset++) {
    if(strchr(setstr, rep->buf[offset]))
      return offset;
  }
  return -1;
}

int mystring::find_last(mystringtmp set) const
{
  const char* setstr = set;
  for(int i = rep->length-1; i >= 0; --i) {
    if(strchr(setstr, rep->buf[i]))
      return i;
  }
  return -1;
}

// return the sub-string ending at 'offset'
mystring mystring::left(size_t offset) const
{
  if(offset > rep->length)
    return mystring();
  else
    return mystring(rep->buf, offset);
}

// return the sub-string starting at 'offset'
mystring mystring::right(size_t offset) const
{
  if(offset >= rep->length)
    return mystring();
  else
    return mystring(rep->buf+offset, rep->length-offset);
}

// return the 'len' characters of the string starting at 'offset'
mystring mystring::sub(size_t offset, size_t len) const
{
  // return right(offset).left(len);
  if(len == 0)
    return mystring();
  else {
    if(len+offset >= rep->length)
      len = rep->length - offset;
    return mystring(rep->buf+offset, len);
  }
}

mystring mystring::lstrip() const
{
  const char* ptr = rep->buf;
  while(*ptr && isspace(*ptr))
    ++ptr;
  return ptr;
}

mystring mystring::rstrip() const
{
  const char* ptr = rep->buf + rep->length - 1;
  while(ptr >= rep->buf && isspace(*ptr))
    --ptr;
  return mystring(rep->buf, ptr-rep->buf+1);
}

mystring mystring::strip() const
{
  const char* start = rep->buf;
  while(*start && isspace(*start))
    ++start;
  const char* end = rep->buf + rep->length - 1;
  while(end >= start && isspace(*end))
    --end;
  return mystring(start, end-start+1);
}

mystring mystring::operator+=(mystringtmp that)
{
  append(that);
  return *this;
}
  

mystring mystring::operator+=(char ch)  
{
  char str[2] = { ch, 0 };
  append(mystringtmp(str, 1));
  return *this;
}
    
mystring operator+(mystringtmp s1, mystringtmp s2)
{
  return mystring(s1, s2);
}
